#!/usr/bin/env python

# Copyright (c) 2014,2016,2022 University of Pennsylvania
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
The command line interface to the SafeDNS zone management module
"""

from __future__ import print_function

__author__ = "John W. O'Brien"
__copyright__ = "Copyright 2014,2016,2022 University of Pennsylvania"
__license__ = "BSD2CLAUSE"
__maintainer__ = "John W. O'Brien"
__email__ = "obrienjw@upenn.edu"
__status__ = "Stable"

import sys
import os.path
import argparse

try:
    import configparser
except ImportError:
    # python 2 compatibility
    import ConfigParser as configparser

from safedns import proxy

rcfile = '~/.safednsrc'

class_map = {
    'white':    proxy.SafeDNSWhitelist,
    'black':    proxy.SafeDNSBlacklist,
}

method_map = {
    'add':      'add',
    'del':      'delete',
    'chk':      'exists',
    'ann':      'annotations',
}

bool_exit_status_map = {
    True:       0,
    False:      1,
}


def parse_args():
    parser = argparse.ArgumentParser(
            description="SafeDNS Zone Management Utility",
    )
    parser.add_argument(
            '-c',
            default=rcfile,
            dest='rcfile',
            help='runtime configuration file (ini-style)',
    )
    parser.add_argument(
            '-ns',
            action='store_true',
            help='operate on a nameserver, not a regular domain name',
    )
    parser.add_argument(
            '-v',
            action='store_true',
            dest='verbose',
            help='emit additional progress and result output',
    )

    subparsers = parser.add_subparsers(
            dest='action',
            help='sub-commands',
    )
    parser_add = subparsers.add_parser(
            'add',
            description='Add a name(server) and optional annotation to zone',
    )
    parser_add.add_argument(
            '--msg',
            type=str,
            help='annotation message',
    )
    subparsers.add_parser(
            'del',
            description='Delete a name(server) and all annotations from zone',
    )
    subparsers.add_parser(
            'chk',
            description='Check for existence of name(server) in zone',
    )
    subparsers.add_parser(
            'ann',
            description='Fetch annotations for a name(server) in zone',
    )

    parser.add_argument(
            'zone',
            choices=['white', 'black', 'feed'],
            help='zone on which to operate',
    )
    parser.add_argument(
            'domain',
            type=str,
            help='domain name for operation',
    )

    return parser.parse_args()


def get_config(rcfile=rcfile):
    config = configparser.ConfigParser(allow_no_value=True)
    config.read(os.path.expanduser(rcfile))
    if 'server' not in config:
        raise RuntimeError(
                'ERROR: Runtime-config file ({}) missing [server] section'
                .format(rcfile)
        )
    return dict(config.items('server'))


def main():
    args = parse_args()
    try:
        server = get_config(args.rcfile)
    except RuntimeError as e:
        print(e.msg)
        sys.exit(1)

    sdns_cls = class_map[args.zone]
    zone = sdns_cls(**server)

    if getattr(args, 'msg', False):
        method_args = (args.domain, args.msg)
    else:
        method_args = (args.domain,)

    try:
        method_name = method_map[args.action]
    except KeyError:
        raise NotImplementedError('I cannot yet do action ' + args.action)
    if args.ns:
        method_name = method_name + 'ns'
    method = getattr(zone, method_name)
    ret = method(*method_args)
    if isinstance(ret, str):
        print(ret)
    elif isinstance(ret, bool):
        if args.verbose:
            print(ret)
        sys.exit(bool_exit_status_map[ret])
    elif isinstance(ret, int):
        sys.exit(ret)


if __name__ == '__main__':
    main()
