#!/usr/bin/env python

# Copyright (c) 2014,2022 University of Pennsylvania
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
A simple Dynamic DNS wrapper module to manage SafeDNS zones
"""

__author__ = "John W. O'Brien"
__copyright__ = "Copyright 2014,2022 University of Pennsylvania"
__license__ = "BSD2CLAUSE"
__maintainer__ = "John W. O'Brien"
__email__ = "obrienjw@upenn.edu"
__status__ = "Stable"

from datetime import datetime

import dns.update
import dns.query
import dns.tsig
import dns.tsigkeyring
import dns.resolver
import dns.zone


class SafeDNSZone(object):
    """
    The base class implementing parameterizable functionality
    """

    # subclasses must set these to useful values
    zone = None
    target = None
    ttl = None

    def __init__(self, host, keyname, secret, algorithm):
        """
        Initialize a SafeDNS zone proxy object

        Parameters
        ----------
        host : str
                hostname of a DNS server that can accept updates
                e.g. 'myserver.example.com'
        keyname : str
                the TSIG key name
                e.g. 'user.tsig.example.com'
        secret : str
                a base64-encoded TSIG secret
                e.g. 'QWgsIGhhISBJIHNlZSB5b3UncmUgdGhlIG5vc2V5IHR5cGUsIGVoPwo='
        algorithm : str
                the name of a supported TSIG algorithm
                e.g. 'hmac-md5'
        """
        self.server_name = host
        self.server_addr = dns.resolver.query(host)[0].address
        self.keyname = keyname
        self.keyring = dns.tsigkeyring.from_text({self.keyname: secret})
        self.keyalgorithm = algorithm
        self.resolver = dns.resolver.Resolver(configure=False)
        self.resolver.nameservers = [self.server_addr]
        self.resolver.use_tsig(self.keyring, self.keyname, self.keyalgorithm)

    def exists(self, name):
        """
        Test whether `name` already appears in this zone

        Parameters
        ----------
        name : str
                the domain name to probe
                e.g. 'somedomain.example.net'

        Returns
        -------
        bool
                True if a `name` CNAME record is in this zone, False otherwise
        """
        try:
            self.resolver.query('{}.{}'.format(name, self.zone), 'CNAME')
        except (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer):
            return False
        return True

    def existsns(self, name):
        """
        Test whether the nameserver `name` already appears in this zone

        Parameters
        ----------
        name : str
                the nameserver name to probe
                e.g. 'ns1.example.net'

        Returns
        -------
        bool
                True if a properly formatted CNAME record based on `name` is
                in this zone, False otherwise
        """
        rpz_name = name + '.rpz-nsdname'
        return self.exists(rpz_name)

    def annotations(self, name):
        """
        Retrieve all notes attached to the entry `name` in this zone.

        Parameters
        ----------
        name : str
                the domain name to probe

        Returns
        -------
        str
                a quasi-formatted block of text of the form

                    "<timestamp0>" "<note0>"
                    "<timestamp1>" "<note1>"
                    ...
                    "<timestampN>" "<noteN>"

                The double quotes are part of the output. If no annotations
                exist, return the empty string. Double quotes are ommitted
                if any note is empty.
        """
        try:
            qname = '_.{}.{}'.format(name, self.zone)
            ans = self.resolver.query(qname, rdtype='TXT')
        except dns.resolver.NXDOMAIN:
            return ''

        ret = ''
        for rd in sorted(ans):
            ret += rd.to_text() + '\n'
        return ret

    def annotationsns(self, name):
        """
        Retrieve all notes attached to the nameserver entry `name` in
        this zone.

        Parameters
        ----------
        name : str
                the nameserver name to probe

        Returns
        -------
        str
                a quasi-formatted block of text of the form

                    "<timestamp0>" "<note0>"
                    "<timestamp1>" "<note1>"
                    ...
                    "<timestampN>" "<noteN>"

                The double quotes are part of the output. If no annotations
                exist, return the empty string. Double quotes are ommitted
                if any note is empty.
        """
        rpz_name = name + '.rpz-nsdname'
        return self.annotations(rpz_name)

    def prep_update(self):
        """
        Prepare, but do not populate or send, a DNS dynamic update query

        Returns
        -------
        dns.update.Update
                this object has been initialized with the zone and TSIG key
                information, but must be populated with any desired
                prerequisites and updates by the caller, then sent
        """
        return dns.update.Update(
                self.zone,
                keyring=self.keyring,
                keyname=self.keyname,
                keyalgorithm=self.keyalgorithm
        )

    def add(self, name, note=None):
        """
        Add `name` to the zone, optionally including a `note`.

        If `name` already exists in this zone, `note` will be appended.

        Parameters
        ----------
        name : str
                the domain name to add
        note : str, optional
                a note to be attached to the name as a single string in a
                TXT record
        """
        u = self.prep_update()
        u.add(name, self.ttl, 'CNAME', self.target)
        u.add('*.{}'.format(name), self.ttl, 'CNAME', self.target)
        # ISO 8601
        formatted_dt = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        if note is not None:
            txt_rdata = '"{}" "{}"'.format(formatted_dt, note)
        else:
            txt_rdata = '"{}"'.format(formatted_dt)
        u.add('_.{}'.format(name), self.ttl, 'TXT', txt_rdata)
        dns.query.tcp(u, self.server_addr, timeout=10)

    def addns(self, name, note=None):
        """
        Add the nameserver `name` to the zone, optionally including a `note`.

        If a nameserver called `name` already exists in this zone, `note`
        will be appended.

        Parameters
        ----------
        name : str
                the nameserver name to add
        note : str, optional
                a note to be attached to the nameserver as a single string
                in a TXT record
        """
        u = self.prep_update()
        rpz_name = name + ".rpz-nsdname"
        u.add(rpz_name, self.ttl, 'CNAME', self.target)
        # ISO 8601
        formatted_dt = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        if note is not None:
            txt_rdata = '"{}" "{}"'.format(formatted_dt, note)
        else:
            txt_rdata = '"{}"'.format(formatted_dt)
        u.add('_.' + rpz_name, self.ttl, 'TXT', txt_rdata)
        dns.query.tcp(u, self.server_addr, timeout=10)

    def delete(self, name):
        """
        Delete `name` from the zone (if it exists), and all attached notes.

        Parameters
        ----------
        name : str
                the domain name to delete
        """
        u = self.prep_update()
        u.delete(name, 'CNAME')
        u.delete('*.{}'.format(name), 'CNAME')
        u.delete('_.{}'.format(name), 'TXT')
        dns.query.tcp(u, self.server_addr, timeout=10)

    def deletens(self, name):
        """
        Delete the nameserver `name` from the zone (if it exists), and all
        attached notes.

        Parameters
        ----------
        name : str
                the nameserver name to delete
        """
        u = self.prep_update()
        rpz_name = name + ".rpz-nsdname"
        u.delete(rpz_name, 'CNAME')
        u.delete('_.' + rpz_name, 'TXT')
        dns.query.tcp(u, self.server_addr, timeout=10)


class SafeDNSWhitelist(SafeDNSZone):
    """
    A proxy class for the SafeDNS whitelist.
    """
    zone = 'safedns_whitelist.'
    ttl = 3600
    target = 'rpz-passthru.'


class SafeDNSBlacklist(SafeDNSZone):
    """
    A proxy class for the SafeDNS blacklist
    """
    zone = 'safedns_blacklist.'
    ttl = 3600
    target = '.'


if __name__ == '__main__':
    pass
